#ifndef TWOTHREE_H
#define TWOTHREE_H

#include <stddef.h>
#include "Item.h"

class ostream;

// TwoThreeNode -- Class to implement a 2-3 Tree
//
class TwoThreeNode {
public:
   enum  {
      MAX_KEYS = 2,
      MAX_SUBTREES = (MAX_KEYS + 1),
      MEDIAN_INDEX = (MAX_SUBTREES / 2)
   };

      // Indices into a subtree array
   enum  dir_t { LEFT = 0, MIDDLE = 1, RIGHT = 2 };

   // ----- Constructors and destructors: 

   TwoThreeNode(ItemType * item1=NULL, ItemType * item2=NULL);
   virtual ~TwoThreeNode(void);

      // Compare this item to the given key and return:
      //   EQ_CMP if the key is equal
      //   MIN_CMP if this key is less than the current node's keys
      //   MAX_CMP if this key is greater than the current node's key
      // Upon return, the parameter "index" will be set as follows:
      //   If the key was found, index will be the index of
      //   key in this node's keyfield array. Otherwise, index
      //   will be the index of the subtree to continue searching
      //   for "key" in.
   cmp_t
   Compare(KeyType key, int & index, cmp_t cmp=EQ_CMP) const;

   // ----- Query attributes:

      // Get this node's key value
   KeyType
   Key(dir_t  dir) const {
      return  (dir > MIDDLE) ? myData[MIDDLE]->Key()
                             : myData[dir]->Key();
   }

      // Get the number of keys stored in this node
   int
   NumKeys() const {
      return  myNumKeys;
   }

      // Get given subtree of this tree
      //
   TwoThreeNode *
   Subtree(dir_t dir) const { return  mySubtree[dir]; }

   // ----- Search/Insert/Delete
   //
   //   NOTE: These are all static functions instead of member functions
   //         because most of them need to modify the given tree root
   //         pointer. If these were instance member functions than
   //         that would correspond to having to modify the 'this'
   //         pointer, which is not allowed in C++. Most of the 
   //         functions that are static and which take an 2-3 tree
   //         pointer as a parameter are static for this reason.
   
      // Look for the given key, return NULL if not found,
      // otherwise return the node.
   static ItemType *
   Search23(KeyType key, TwoThreeNode * root, cmp_t cmp=EQ_CMP);

      // Search is an alias for Search23
   static ItemType *
   Search(KeyType key, TwoThreeNode * root, cmp_t cmp=EQ_CMP) {
      return  Search23(key, root, cmp);
   }

      // Insert the given key, return NULL if it was inserted,
      // otherwise return the existing node  with the same key.
   static ItemType *
   Insert23(ItemType * item, TwoThreeNode * & root);

      // Insert is an alias for Insert23
   static ItemType *
   Insert(ItemType * item, TwoThreeNode * & root) {
      return  Insert23(item, root);
   }

      // Delete the given key from the tree. Return the corresponding
      // node, or return NULL if it was not found.
   static ItemType *
   Delete23(KeyType key, TwoThreeNode * & root, cmp_t cmp=EQ_CMP);

      // Delete is an alias for Delete23
   static ItemType *
   Delete(KeyType key, TwoThreeNode * & root, cmp_t cmp=EQ_CMP) {
      return  Delete23(key, root, cmp);
   }

   // Verification 

      // Return the height of this tree
   int
   Height() const;

      // Verify this tree is a valid 2-3 tree, return TRUE if it is,
      // return FALSE otherwise
   int
   Check23() const;

      // Check is an alias for Check23
   int
   Check() const {
      return  Check23();
   }

      // If you want to provide your own allocation scheme than simply
      // #define the preprocessor manifest constant named CUSTOM_ALLOCATE
      // and make sure you provide and link with your own overloaded
      // versions of operators "new" and "delete" for this class.
#ifdef CUSTOM_ALLOCATE
   void *
   operator  new(size_t);

   void
   operator  delete(void *);
#endif  /* CUSTOM_ALLOCATE */


private:

   // ----- Private data

   ItemType     * myData[MAX_KEYS];  // Data fields
   TwoThreeNode * mySubtree[MAX_SUBTREES];   // Pointers to subtrees
   short          myNumKeys;         // Number of keys

      // Reset all subtrees to null
   void
   Reset(void) {
      mySubtree[LEFT] = mySubtree[MIDDLE] = mySubtree[RIGHT] = NULL ;
   }

   // ----- Routines that do the *real* insertion/deletion

      // Insert the given key into the given tree. Return the node if
      // it already exists. Otherwise return NULL to indicate that
      // the key was successfully inserted.  Upon return, the "overflow"
      // parameter will be NULL if the key was successfully inserted
      // at this level without causing an node overflow. Otherwise,
      // "overflow" will be a pointer to the new node that was created
      // as a result of splitting the overfull node, and key will be reset
      // to the value of the key that wouldnt fit.
   static ItemType *
   Insert23(ItemType * & item, TwoThreeNode * & root,
                               TwoThreeNode * & overflow);

      // Delete the given key from the given tree. Return NULL if the
      // key is not found in the tree. Otherwise return a pointer to the
      // node that was removed from the tree. Upon return, the "underflow"
      // parameter will be NULL if the key was successfully deleted
      // at this level without causing an node underflow. Otherwise,
      // "overflow" will be non-NULL and ???
   static ItemType *
   Delete23(KeyType key, TwoThreeNode * & root,
                         TwoThreeNode * & overflow, cmp_t cmp=EQ_CMP);

   // Routines for restructuring

      // Split the current node into to two nodes. Upon return, the splitting
      // (median) item gets put into "item", and the new node is pointed to
      // the "overflow"
   void
   SplitNode(ItemType * & item, int  index, TwoThreeNode * & overflow);

      // Internally insert the given data into this node at position
      // index (overflowing into "overflow" if necessary). Reference
      // parameters "item" and "overflow" have the same semantics 
      // as for Insert23().
   void
   InternalInsert(ItemType * & item, int  index, TwoThreeNode * & overflow);

      // Merge the node "underflow" (which is the index-th subtree of
      // this node) with its given sibling.
   void
   MergeNodes(int  index, TwoThreeNode * underflow, TwoThreeNode * sibling);

      // The index-th key has too few data items, redistribute by either
      // borrowing from, or merging with a sibling.
   void
   Redistribute(int  index);

      // Internally delete the index-th data item from this node.
   void
   InternalDelete(int  index);

private:
   // Disallow copying and assignment
   TwoThreeNode(const TwoThreeNode &);
   TwoThreeNode & operator=(const TwoThreeNode &);

};


   // Class TwoThreeTree is a simple container object to "house" a TwoThreeNode
   // that represents the root-node of and TwoThreeTree. Most of the member
   // functions simply delegate to the root TwoThreeNode.
class TwoThreeTree {
private:
      // Disallow copying and assingment
   TwoThreeTree(const TwoThreeTree &);
   TwoThreeTree & operator=(const TwoThreeTree &);

      // Member data
   TwoThreeNode * myRoot;   // The root of the tree
   static const char  ourKind[];   // What kind of tree this is

public:
      // Constructor and destructor
   TwoThreeTree() : myRoot(NULL) {};
   ~TwoThreeTree() { if (myRoot)  delete myRoot; }

      // Print the tree kind
   const char *
   Kind() const { return ourKind; }

      // Dump the tree to the given output stream
   void
   DumpTree(ostream & os) const;

      // See if the tree is empty
   int
   IsEmpty() const {
      return  (myRoot == NULL);
   }

      // Search, Insert, Delete, and Check
   ItemType *
   Search(KeyType key, cmp_t cmp=EQ_CMP) {
      return  TwoThreeNode::Search23(key, myRoot, cmp);
   }

   ItemType *
   Insert(ItemType * item) {
      return  TwoThreeNode::Insert23(item, myRoot);
   }

   ItemType *
   Delete(KeyType key, cmp_t cmp=EQ_CMP) {
      return  TwoThreeNode::Delete23(key, myRoot, cmp);
   }

   int
   Check() const {
      return  (myRoot) ? myRoot->Check23() : 1;
   }
};

#endif  /* TWOTHREE_H */
